/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

//! \author Vijay Pradeep

#include "camera_offsetter/generic_offsetter.h"
#include "sensor_msgs/Image.h"
#include "sensor_msgs/CameraInfo.h"

using namespace std;

namespace camera_offsetter
{

class MonoOffsetter : public GenericOffsetter
{
public:
  MonoOffsetter()
  {
    ros::NodeHandle nh_private("~");

    cam_name_ = nh_.resolveName("camera");
    if (cam_name_ == "camera")
      ROS_WARN("camera was not remapped. Are you sure your camera namespace is 'camera'?");

    source_nh_ = ros::NodeHandle(cam_name_);
    offset_nh_ = ros::NodeHandle(cam_name_ + "_offset");

    vector<string> image_topics;

    bool offset_raw = false;
    nh_private.getParam("use_raw", offset_raw);
    if (offset_raw)
      image_topics.push_back("image_raw");

    bool offset_rect = false;
    nh_private.getParam("use_rect", offset_rect);
    if (offset_rect)
      image_topics.push_back("image_rect");

    bool offset_rect_color = false;
    nh_private.getParam("use_rect_color", offset_rect_color);
    if (offset_rect_color)
      image_topics.push_back("image_rect_color");

    for (size_t i=0; i < image_topics.size(); i++)
    {
      ros::Publisher  pub = offset_nh_.advertise<sensor_msgs::Image>(image_topics[i], 1);

      ros::SubscribeOptions ops;

      boost::function< void(const sensor_msgs::ImageConstPtr&) > f = boost::bind(&MonoOffsetter::imageCb, this, _1, pub);
      ros::Subscriber sub = source_nh_.subscribe(image_topics[i], 1, f);
      image_subs_.push_back(sub);
      image_pubs_.push_back(pub);
    }

    info_sub_  = source_nh_.subscribe("camera_info", 1, &MonoOffsetter::infoCb,  this);
    info_pub_  = offset_nh_.advertise<sensor_msgs::CameraInfo>("camera_info", 1);
  }

  void imageCb(const sensor_msgs::ImageConstPtr& msg, ros::Publisher image_pub)
  {
    sensor_msgs::Image next_image = *msg;
    next_image.header.frame_id = msg->header.frame_id + frame_suffix_;
    image_pub.publish(next_image);

    publishTransform(msg->header.stamp, next_image.header.frame_id, msg->header.frame_id);
  }

  void infoCb(const sensor_msgs::CameraInfoConstPtr& msg)
  {
    sensor_msgs::CameraInfo next_info = *msg;
    next_info.header.frame_id = msg->header.frame_id + frame_suffix_;
    info_pub_.publish(next_info);
  }

private:
  ros::NodeHandle nh_;
  ros::NodeHandle source_nh_;
  ros::NodeHandle offset_nh_;

  string cam_name_;

  vector<ros::Subscriber> image_subs_;
  vector<ros::Publisher>  image_pubs_;

  ros::Subscriber info_sub_;
  ros::Publisher  info_pub_;
} ;

}

using namespace camera_offsetter;

int main(int argc, char** argv)
{
  ros::init(argc, argv, "mono_offsetter");

  MonoOffsetter offsetter;

  ros::spin();
}
